#include <gio/gio.h>
#include <glib.h>
#include <gcrypt.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/if.h>
#include <unistd.h>
#include <netinet/in.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>

#include "identifier.h"

static char* ofono_name = "org.ofono";
static gint fail_next_get_mac = 0;
static gint fail_next_get_uuid = 0;

void whoopsie_identifier_set_ofono_name (char *name) {
    ofono_name = name;
}

void
whoopsie_hex_to_char (char* buf, const char *str, int len)
{
    char* p = NULL;
    int i = 0;

    g_return_if_fail (buf);
    g_return_if_fail (str);

    p = buf;
    for (i = 0; i < len; i++) {
        snprintf(p, 3, "%02x", (unsigned char) str[i]);
        p += 2;
    }
    buf[2*len] = 0;
}

void
whoopsie_identifier_get_mac_address (char** res, GError** error)
{
    struct ifreq ifr;
    struct ifconf ifc;
    char buf[1024];
    gboolean success = FALSE;
    int sock;
    struct ifreq* it;
    int i = 0;

    g_return_if_fail (res);

    if (g_atomic_int_compare_and_exchange (&fail_next_get_mac, WHOOPSIE_FAIL_GENERIC, 0)) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), WHOOPSIE_FAILED_BY_REQUEST,
                     "Failed by tester request");
        return;
    }

    sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
    if (sock == -1) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "Could not create socket.");
        return;
    }

    ifc.ifc_len = sizeof(buf);
    ifc.ifc_buf = buf;
    if (ioctl(sock, SIOCGIFCONF, &ifc) == -1) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "Could not get list of interface addresses.");
        close (sock);
        return;
    }

    it = ifc.ifc_req;

    if (!g_atomic_int_compare_and_exchange (&fail_next_get_mac, WHOOPSIE_FAIL_MAC_SIOCGIFFLAGS, 0)) {
        for (i = ifc.ifc_len / sizeof(struct ifreq); --i >= 0; it++) {
            strcpy(ifr.ifr_name, it->ifr_name);
            if (ioctl(sock, SIOCGIFFLAGS, &ifr) == 0) {
                if (! (ifr.ifr_flags & IFF_LOOPBACK)) { // don't count loopback
                    if (ioctl(sock, SIOCGIFHWADDR, &ifr) == 0) {
                        success = TRUE;
                        break;
                    }
                }
            }
        }
    }

    close (sock);

    if (!success) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "Unable to find a hardware address");
        return;
    }

    *res = g_malloc (18);
    sprintf (*res, "%02x:%02x:%02x:%02x:%02x:%02x",
             (unsigned char) ifr.ifr_hwaddr.sa_data[0],
             (unsigned char) ifr.ifr_hwaddr.sa_data[1],
             (unsigned char) ifr.ifr_hwaddr.sa_data[2],
             (unsigned char) ifr.ifr_hwaddr.sa_data[3],
             (unsigned char) ifr.ifr_hwaddr.sa_data[4],
             (unsigned char) ifr.ifr_hwaddr.sa_data[5]);
}

void
whoopsie_identifier_get_system_uuid (char** res, GError** error)
{
    int fp;

    g_return_if_fail (res);

    if (g_atomic_int_compare_and_exchange (&fail_next_get_uuid, WHOOPSIE_FAIL_GENERIC, 0)) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), WHOOPSIE_FAILED_BY_REQUEST,
                     "Failed by tester request");
        return;
    }

    fp = open ("/sys/class/dmi/id/product_uuid", O_RDONLY);
    if (fp < 0) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "Could not open the product uuid file.");
        return;
    }
    *res = g_malloc (37);
    if (read (fp, *res, 36) == 36)
        (*res)[36] = '\0';
    else
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "Got an unexpected length reading the product_uuid.");
    close (fp);
}

void
whoopsie_identifier_sha512 (char* source, char* res, GError** error)
{
    int md_len;
    gcry_md_hd_t sha512 = NULL;
    unsigned char* id = NULL;

    g_return_if_fail (source);
    g_return_if_fail (res);

    if (!gcry_check_version (GCRYPT_VERSION)) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "libcrypt version mismatch.");
        return;
    }
    gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
    md_len = gcry_md_get_algo_dlen(GCRY_MD_SHA512);
    if (md_len != HASHLEN / 2) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                 "Received an incorrect size for the SHA512 message digest");
        return;
    }
    if (gcry_md_open (&sha512, GCRY_MD_SHA512, 0) || sha512 == NULL) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
             "Failed to create a SHA512 message digest for the product_uuid.");
        return;
    }
    gcry_md_write (sha512, source, strlen (source));
    gcry_md_final (sha512);
    id = gcry_md_read (sha512, GCRY_MD_SHA512);
    if (id == NULL) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
             "Failed to read the SHA512 message digest for the product_uuid.");
        gcry_md_close (sha512);
        return;
    }
    whoopsie_hex_to_char (res, (const char*)id, md_len);
    gcry_md_close (sha512);
}

void whoopsie_identifier_append_imei (char** result, GError** error)
{
    GVariant* res;
    GError* err = NULL;

    GDBusConnection* con = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &err);

    if (err != NULL) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "Unable to connect to the bus: %s", err->message);
        return;
    }

    res = g_dbus_connection_call_sync (con, ofono_name, "/",
                                       "org.ofono.Manager", "GetModems",
                                       NULL, G_VARIANT_TYPE ("(a(oa{sv}))"),
                                       G_DBUS_CALL_FLAGS_NONE,
                                       -1, NULL, &err);

    if (err != NULL) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "Unable to call ofono's GetModems: %s", err->message);
        return;
    }
    g_object_unref (con);

    /* res is an a(oa{sv}) */
    GVariant* var1 = g_variant_get_child_value (res, 0);
    g_variant_unref (res);
    /* but the first array can be empty */
    if (g_variant_n_children (var1) < 1) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "No modems in GetModems response");
        return;
    }
    GVariant* var2 = g_variant_get_child_value (var1, 0);
    g_variant_unref (var1);
    GVariant* var3 = g_variant_get_child_value (var2, 1);
    g_variant_unref (var2);
    GVariant* var4 = g_variant_lookup_value (var3, "Serial", G_VARIANT_TYPE ("s"));
    g_variant_unref (var3);
    if (var4 == NULL) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "No Serial key in GetModems response");
        return;
    }
    gsize imei_len;
    const gchar* imei = g_variant_get_string(var4, &imei_len);
    if (imei_len < 15 || imei_len > 16) {
        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
                     "GetModems' Serial key is not an IMEI (has length %zu)", imei_len);
        return;
    }

    size_t old_len = *result ? strlen(*result) : 0;

    *result = g_realloc (*result, old_len + imei_len + 1);
    strcpy (*result + old_len, imei);

    g_variant_unref(var4);
}

void
whoopsie_identifier_generate (char** res, GError** error)
{
    char* identifier = NULL;

    g_return_if_fail (res);

    whoopsie_identifier_get_system_uuid (&identifier, error);
    if ((!error || !(*error)) && identifier)
        goto out;

    if (error && *error) {
        g_error_free (*error);
        *error = NULL;
    }

    whoopsie_identifier_get_mac_address (&identifier, error);
    if ((!error || !(*error)) && identifier) {
        whoopsie_identifier_append_imei (&identifier, error);
        if (error && *error) {
            g_error_free (*error);
            *error = NULL;
        }

        goto out;
    }

    return;

out:
    *res = g_malloc (HASHLEN + 1);
    whoopsie_identifier_sha512 (identifier, *res, error);
    g_free (identifier);
}

void whoopsie_identifier_fail_next_get_mac (gint how)
{
    g_atomic_int_set (&fail_next_get_mac, how);
}

void whoopsie_identifier_fail_next_get_uuid ()
{
    g_atomic_int_set (&fail_next_get_uuid, 1);
}
